import os
import logging
import tempfile

from ..elf import get_shobj_deps, is_dynamic, LddError
from ..utils import make_executable, mkdirs_for

def process_pyinstaller_archive(sx):
    # See utils/cliutils/archive_viewer.py

    # If PyInstaller is not installed, do nothing
    try:
        from PyInstaller.archive.readers import CArchiveReader, NotAnArchiveError
    except ImportError:
        return

    # Attempt to open the program as PyInstaller archive
    try:
        pyi_ar = CArchiveReader(sx.orig_prog)
    except:
        # Silence all PyInstaller exceptions here
        return
    logging.info("Opened PyInstaller archive!")

    with PyInstallHook(sx, pyi_ar) as h:
        h.process()



class PyInstallHook:
    def __init__(self, sx, pyi_archive):
        self.sx = sx
        self.pyi_ar = pyi_archive

        self.tmpdir = tempfile.TemporaryDirectory(prefix='staticx-pyi-')


    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.tmpdir.cleanup()


    def process(self):
        binaries = self._extract_binaries()
        for binary in binaries:
            # These could be Python libraries, shared object dependencies, or
            # anything else a user might add via `binaries` in the .spec file.
            self._add_required_deps(binary)


    def _extract_binaries(self):
        result = []

        for n, item in enumerate(self.pyi_ar.toc.data):
            (dpos, dlen, ulen, flag, typcd, name) = item

            # Only process binary files
            # See xformdict in PyInstaller.building.api.PKG
            if typcd != 'b':
                continue

            # Extract it to a temporary location
            _, data = self.pyi_ar.extract(n)
            tmppath = os.path.join(self.tmpdir.name, name)
            logging.debug("Extracting to {}".format(tmppath))
            mkdirs_for(tmppath)
            with open(tmppath, 'wb') as f:
                f.write(data)

            # We can't use yield here, because we need all of the libraries to be
            # extracted prior to running ldd (see #61)
            result.append(tmppath)

        return result


    def _add_required_deps(self, lib):
        """Add dependencies of lib to staticx archive"""

        # Verify this is a shared library
        if not is_dynamic(lib):
            # It's okay if there's a static executable in the PyInstaller
            # archive. See issue #78
            return

        # Silence "you do not have execution permission" warning from ldd
        make_executable(lib)

        # Check for RPATH/RUNPATH, but only "dangerous" values and let
        # "harmless" values pass (e.g. "$ORIGIN/cffi.libs")
        self.sx.check_library_rpath(lib, dangerous_only=True)
        # Unfortunately, there's no easy way to fix an UnsupportedRunpathError
        # here, because staticx is not about to to modify the library and
        # re-pack the PyInstaller archive itself.

        # Try to get any dependencies of this file
        try:
            deps = get_shobj_deps(lib, libpath=[self.tmpdir.name])
        except LddError as e:
            # In certain cases, ldd might get upset about binary files
            # (probably added by the user via .spec file). This can happen
            # e.g., if a dynamically-linked musl-libc application is included.
            # There's a reasonable chance this won't run, but it's not really
            # staticx's problem, so warn the user and go on.
            logging.warning(e)
            return


        # Add any missing libraries to our archive
        for deppath in deps:
            dep = os.path.basename(deppath)

            if self.pyi_ar.toc.find(dep) != -1:
                logging.debug("{} already in pyinstaller archive".format(dep))
                continue

            self.sx.add_library(deppath, exist_ok=True)



from .pyinstaller import process_pyinstaller_archive
from .glibc import process_glibc_prog

hooks = [
    process_pyinstaller_archive,
    process_glibc_prog,
]


def run_hooks(sx):
    for hook in hooks:
        hook(sx)



from SCons.Action import Action

def read_nsswitch_conf(path):
    with open(path, 'r') as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            dbname, svcs = line.split(':')
            dbname = dbname.strip()
            svcs = svcs.strip().split()
            yield dbname, svcs

def _gen_nsswitch_conf_h(target, source, env):
    conf = read_nsswitch_conf(source[0].get_path())
    with open(target[0].get_path(), 'w') as tgtf:
        for dbname, svcs in conf:
            tgtf.write('NSSWITCH_CONF("{}", "{}")\n'.format(dbname, ' '.join(svcs)))


def _strfunc(target, source, env):
    return "Creating '%s'" % target[0]

def NsswitchConfH(env, target, source):
    return env.Command(
        action = Action(_gen_nsswitch_conf_h, _strfunc),
        target = target,
        source = source,
    )

def GetNsswitchLibs(env, source):
    source = env.File(source)
    libs = set()
    conf = read_nsswitch_conf(source.srcnode().abspath)
    for dbname, svcs in conf:
        libs.update('libnss_'+s for s in svcs)
    return list(libs)

def generate(env):
    env.AddMethod(NsswitchConfH)
    env.AddMethod(GetNsswitchLibs)
    return True

def exists(env):
    return True



import pkg_resources
from .utils import copy_fileobj_to_tempfile

def locate_asset(name, debug):
    mode = 'debug' if debug else 'release'
    path = '/'.join(('assets', mode, name))
    try:
        return pkg_resources.resource_stream(__name__, path)
    except FileNotFoundError:
        raise KeyError("Asset not found: {!r} (mode={!r})".format(name, mode))


def copy_asset_to_tempfile(assetname, debug, **kwargs):
    with locate_asset(assetname, debug) as fsrc:
        return copy_fileobj_to_tempfile(fsrc, **kwargs)



from SCons.Util import is_List

def ShlibStubGen(env, target, symbols):
    """Generate a shared library stub

    This stub is used to facilitate dynamic linking without including
    occasionally problematic symbol versioning information.

    Parameters:
        env:        Environment
        target:     Target shared object name (e.g. libc.so.6)
        symbols:    Symbol list to include in stub
    """
    # Inspired by https://stackoverflow.com/a/21059674

    if is_List(target):
        raise UserError("target must be a single string or node")
    target = env.fs.File(target)

    lines = ['// Auto-generated stub']
    for sym in symbols:
        lines += [
            'void {}(void);'.format(sym),       # for strict prototypes
            'void {}(void) {{}}'.format(sym),
            '',
        ]

    csrc = env.Textfile(
        target = target.name + '_stub.c',
        source = lines,
    )

    stub = env.SharedLibrary(
        target = target,
        source = [csrc],
        CCFLAGS = env['CCFLAGS'] + [
            '-fno-builtin',
        ],
        LINKFLAGS = env['LINKFLAGS'] + [
            '-Wl,-soname=${TARGET.name}',
            '-nostdlib',
        ],
        SHLIBSUFFIX='',
    )
    return stub


def generate(env):
    env.Tool('textfile')
    env.AddMethod(ShlibStubGen)

def exists(env):
    return True



import tarfile
import logging
import lzma
from os.path import basename

from .bcjfilter import get_bcj_filter_arch
from .utils import get_symlink_target, make_mode_executable
from .constants import *
from .errors import *


def get_bcj_filter():
    arch = get_bcj_filter_arch()
    if not arch:
        return None, ''

    # Get the lzma module constant name and value
    filt_name = 'FILTER_' + arch
    filt = getattr(lzma, filt_name)

    return filt, filt_name


def get_xz_filters():
    filters = []

    # Get a BCJ filter for the current architecture
    bcj_filter, bcj_filter_name = get_bcj_filter()
    if bcj_filter:
        logging.info("Using XZ BCJ filter {}".format(bcj_filter_name))
        filters.append(dict(id=bcj_filter))

    # The last filter in the chain must be a compression filter.
    filters.append(dict(id=lzma.FILTER_LZMA2))
    return filters

class SxArchive:
    def __init__(self, fileobj, mode, compress):
        """Create a staticx archive

        Parameters:
        fileobj:    File object for the archive (not closed by this class)
        mode:       Mode: 'r' (for reading) or 'w' (for writing)
        compress:   Boolean: use xz compression or not
        """
        # Keep original fileobj arg for consumer convenience only, never closed
        self.fileobj = fileobj
        self.xzf = None

        if compress:
            self.xzf = lzma.open(
                filename = fileobj,
                mode = mode,
                format = lzma.FORMAT_XZ,

                # Use CRC32 instead of CRC64 (FORMAT_XZ default)
                # Otherwise, enable XZ_USE_CRC64 in libxz/xz_config.h
                check = lzma.CHECK_CRC32,

                filters = get_xz_filters(),
            )

            fileobj = self.xzf

        # Our embedded libtar only supports older GNU format (not new PAX format)
        self.tar = tarfile.open(fileobj=fileobj, mode=mode, format=tarfile.GNU_FORMAT)

    def __enter__(self):
        return self

    def __exit__(self, *excinfo):
        self.close()

    def close(self):
        # Don't touch self.fileobj here

        if self.tar:
            self.tar.close()
            self.tar = None

        if self.xzf:
            self.xzf.close()
            self.xzf = None


    def add_symlink(self, name, target):
        """Add a symlink to the archive"""
        t = tarfile.TarInfo()
        t.type = tarfile.SYMTYPE
        t.name = name
        t.linkname = target

        self.tar.addfile(t)

    def add_fileobj(self, name, fileobj):
        logging.info("Adding {}".format(name))
        tarinfo = self.tar.gettarinfo(arcname=name, fileobj=fileobj)
        tarinfo.mode = make_mode_executable(tarinfo.mode)
        self.tar.addfile(tarinfo, fileobj)

    def add_program(self, path, name):
        """Add user program to the archive

        This adds the user program to the archive using its original filename.
        Additionally, a symlink to the program is created with a fixed name to
        enable the bootloader to identify the program to execute.

        Parameters:
        path:   The path to the program to add
        name:   The original filename of the program

        Should only be called once. TODO: Enforce this.
        """
        def make_exec(tarinfo):
            tarinfo.mode = make_mode_executable(tarinfo.mode)
            return tarinfo

        logging.info("Adding {} as {}".format(path, name))
        self.tar.add(path, arcname=name, filter=make_exec)

        # Store a link to the program so the bootloader knows what to execute
        self.add_symlink(PROG_FILENAME, name)

    def add_file(self, path, arcname=None):
        self.tar.add(path, arcname=arcname)

    def add_interp_symlink(self, interp):
        """Add symlink for ld.so interpreter"""
        self.add_symlink(INTERP_FILENAME, basename(interp))



import platform

# NOTE: This is also used by libxz/SConscript

def get_bcj_filter_arch():
    """
    Get an appropriate BCJ filter for the current architecture.

    Returns just the architecture part of the BCJ filter name.
    This can be prepended with FILTER_ for a Python lzma module constant,
    or XZ_DEC_ for an XZ Embedded decoder macro.
    """
    # It's possible that the code being compressed might not match that of the
    # current Python interpreter. In that case the problem becomes much more
    # complicated, as we have to inspect the files to make a determination.
    # This approach should be "good enough"; the worst case scenario is only a
    # slightly worse compression ratio.

    machine = platform.machine()

    if machine in ('i386', 'i686', 'x86_64'):
        return 'X86'

    if machine == 'ia64':
        return 'IA64'

    if machine.startswith('arm'):   # arm, armv8b, etc.
        return 'ARM'

    # TODO: 'ARMTHUMB'

    if machine.startswith('ppc'):
        return 'POWERPC'

    if machine.startswith('sparc'):
        return 'SPARC'

    return None



import argparse
import sys
import logging

from .api import generate
from .errors import Error
from .version import __version__

def parse_args():
    DEFAULT_LOGLEVEL = 'WARNING'

    ap = argparse.ArgumentParser(prog='staticx')

    # Positional arguments
    ap.add_argument('prog',
            help = 'Input program to bundle')
    ap.add_argument('output',
            help = 'Output path')

    # Operational options
    ap.add_argument('-l', dest='libs', action='append',
            help = 'Add additional libraries (absolute paths)')
    ap.add_argument('--strip', action='store_true',
            help = 'Strip binaries before adding to archive (reduces size)')
    ap.add_argument('--no-compress', action='store_true',
            help = "Don't compress the archive (increases size)")

    # Special / output-related options
    ap.add_argument('-V', '--version', action='version',
            version = '%(prog)s ' + __version__)
    ap.add_argument('--loglevel', choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
            help = 'Set the logging level (default: {})'.format(DEFAULT_LOGLEVEL))

    # Hidden arguments (for development / testing)
    ap.add_argument('--debug', action='store_true',
            help = argparse.SUPPRESS)

    args = ap.parse_args()

    if args.loglevel is None:
        args.loglevel = 'DEBUG' if args.debug else DEFAULT_LOGLEVEL

    return args


def main():
    args = parse_args()
    logging.basicConfig(level=args.loglevel)

    try:
        generate(args.prog, args.output,
                libs = args.libs,
                strip = args.strip,
                compress = not args.no_compress,
                debug = args.debug,
                )
    except Error as e:
        if args.debug:
            raise

        print("staticx: " + str(e))
        sys.exit(2)

if __name__ == '__main__':
    main()



import os
import errno
import shutil
from collections.abc import Iterable
from tempfile import NamedTemporaryFile
from .errors import *

def make_mode_executable(mode):
    mode |= (mode & 0o444) >> 2    # copy R bits to X
    return mode


def make_executable(path):
    mode = os.stat(path).st_mode
    mode = make_mode_executable(mode)
    os.chmod(path, mode)

def get_symlink_target(path):
    dirpath = os.path.dirname(os.path.abspath(path))
    return os.path.join(dirpath, os.readlink(path))

def move_file(src, dst):
    if os.path.isdir(dst):
        raise DirectoryExistsError(dst)
    shutil.move(src, dst)


def mkdirs_for(filename):
    dirname = os.path.dirname(filename)
    os.makedirs(dirname, exist_ok=True)


def copy_to_tempfile(srcpath, **kwargs):
    with open(srcpath, 'rb') as fsrc:
        fdst = copy_fileobj_to_tempfile(fsrc, **kwargs)

    shutil.copystat(srcpath, fdst.name)
    return fdst


def copy_fileobj_to_tempfile(fsrc, **kwargs):
    fdst = NamedTemporaryFile(**kwargs)
    shutil.copyfileobj(fsrc, fdst)
    fdst.flush()
    fdst.seek(0)
    return fdst


def which_exec(name, env=None):
    for path in os.get_exec_path(env=env):
        xp = os.path.join(path, name)
        if os.access(xp, os.X_OK):
            return xp
    return None


def is_iterable(x):
    """Returns true if x is iterable but not a string"""
    return isinstance(x, Iterable) and not isinstance(x, str)

def coerce_sequence(x, t=list):
    if not is_iterable(x):
        x = [x]
    return t(x)

_SENTINEL = object()
_NO_DEFAULT = object()

def single(iterable, key=None, default=_NO_DEFAULT):
    """Returns a single item from iterable

    Arguments:
      iterable: Iterable sequence from which item is taken
      key:      Optional function to select item
      default:  Optional default value to return if no items match,
                otherwise KeyError is raised
    Returns:
      A single item from iterable matching key (if given)
    Raises:
      KeyError if multiple items in sequence match key
      KeyError if no items in sequence match key and default is not given
    """

    if key is None:
        key = lambda _: True

    result = _SENTINEL

    for i in iterable:
        if key(i):
            if result is not _SENTINEL:
                raise KeyError("Multiple items match key")
            result = i

    if result is not _SENTINEL:
        return result

    if default is not _NO_DEFAULT:
        return default

    raise KeyError("No items match key")



def CheckNSS(context):
    context.Message('Checking for NSS...')
    ok, output = context.TryRun(
        extension='.c',
        text="""
#include <nss.h>
int main(void)
{
    __nss_configure_lookup("passwd", "files");
    return 0;
}
""")
    context.Result(ok)
    return ok


custom_tests = dict(
    CheckNSS = CheckNSS,
)



from ..assets import copy_asset_to_tempfile
from ..errors import InternalError
from ..elf import open_elf, get_shobj_deps, patch_elf
from elftools.elf.gnuversions import GNUVerNeedSection
import logging

LIBNSSFIX = 'libnssfix.so'

def process_glibc_prog(sx):
    if not is_linked_against_glibc(sx.orig_prog):
        return

    try:
        nssfix = copy_asset_to_tempfile(LIBNSSFIX, debug=sx.debug,
                prefix='libnssfix-', suffix='.so')
    except KeyError:
        raise InternalError("GLIBC binary detected but libnssfix.so not available")

    # Make the user program depend on libnssfix.so
    patch_elf(sx.tmpprog, add_needed=LIBNSSFIX)

    # Add libnssfix.so and its dependencies to the archive.
    # These include the configured libnss_*.so "service" libs and their
    # dependencies.
    #
    # TODO: Ideally staticx.api.generate_archive() would handle this for us,
    # since we added the dependency on libnssfix to the user program above.
    # Even though dependency discovery runs after this hook, it necessarily
    # operates on the *original* executable and not the copied/modified one,
    # (see dfa201b07e) so it doesn't see changes made here.
    with nssfix:
        # TODO: Don't use sxar
        sx.sxar.add_fileobj(LIBNSSFIX, nssfix)
        for libpath in get_shobj_deps(nssfix.name):
            sx.add_library(libpath, exist_ok=True)


def is_linked_against_glibc(prog):
    with open_elf(prog) as elf:
        sec = elf.get_single_section(GNUVerNeedSection)
        if not sec:
            return False
        for verneed, vernaux_iter in sec.iter_versions():
            if not verneed.name.startswith('libc.so'):
                continue
            for vernaux in vernaux_iter:
                if vernaux.name.startswith('GLIBC_'):
                    logging.debug("Program linked with GLIBC: Found {} {}".format(
                        verneed.name, vernaux.name))
                    return True
    return False




class Error(Exception):
    """Base type for all exceptions raised by staticx"""
    pass

class InternalError(Error):
    """Something internal to staticx went wrong"""
    pass

class ToolError(Error):
    """An external tool indicated an error"""
    def __init__(self, program, message=None):
        super().__init__(message or "'{}' failed".format(program))
        self.program = program

class MissingToolError(Error):
    """A required external tool is missing"""
    def __init__(self, program, package):
        super().__init__("Couldn't find '{}'. Is '{}' installed?".format(
            program, package))
        self.program = program
        self.package = package

class InvalidInputError(Error):
    """Input provided by the user is invalid"""
    pass


class UnsupportedDynTagError(InvalidInputError):
    """A library uses an unsupported dynamic entry

    See https://github.com/JonathonReinhart/staticx/issues/172
    """
    def __init__(self, libpath, value):
        super().__init__("{} uses unsupported {} ({!r}).\n"
                "See https://github.com/JonathonReinhart/staticx/issues/188"
                .format(libpath, self.tag, value))

class UnsupportedRpathError(UnsupportedDynTagError):
    tag = 'DT_RPATH'

class UnsupportedRunpathError(UnsupportedDynTagError):
    tag = 'DT_RUNPATH'


class FormatMismatchError(Error):
    pass


class ArchiveError(Error):
    """Base type for exceptions raised from archive.py"""
    pass

class LibExistsError(ArchiveError):
    """Given library already exists in archive"""
    def __init__(self, lib):
        super().__init__(
                "Library '{}' already exists in archive".format(lib))
        self.lib = lib


class DirectoryExistsError(Error):
    """A given directory already exists"""
    def __init__(self, path):
        super().__init__(
                "{}: is a directory".format(path))
        self.path = path



from tempfile import NamedTemporaryFile, mkdtemp
import tarfile
import os

from .constants import *
from .elf import *
from .utils import *

class ArchiveError(Exception):
    pass

def open_archive(archive):
    f = NamedTemporaryFile(prefix='staticx-archive-', suffix='.tar')
    elf_dump_section(archive, ARCHIVE_SECTION, f.name)

    size = os.stat(f.name).st_size
    if size == 0:
        raise ArchiveError("{} does not appear to contain a staticx archive section"
                .format(archive))
    return tarfile.open(fileobj=f, mode='r', format=tarfile.GNU_FORMAT)


def main():
    import argparse
    ap = argparse.ArgumentParser(
            description='StaticX archive extractor/browser')
    ap.add_argument('archive',
            help="StaticX-generated executable (archive) to extract")
    ap.add_argument('outdir', nargs='?',
            help="Optional output directory into which archive is to be extracted")
    ap.add_argument('-v', '--verbose', action='store_true',
            help="Verbose output")
    args = ap.parse_args()

    try:
        ar = open_archive(args.archive)
    except ArchiveError as e:
        raise SystemExit(str(e))

    with ar:
        if args.outdir:
            ar.extractall(path=args.outdir)
        else:
            ar.list(verbose=args.verbose)

if __name__ == '__main__':
    main()



import subprocess
import sys
import re
import locale
import logging
import errno
import os
from pprint import pformat

import elftools
from elftools.elf.elffile import ELFFile
from elftools.elf.dynamic import DynamicSegment
from elftools.common.exceptions import ELFError

from .errors import *
from .utils import coerce_sequence, single, which_exec


def verify_tools():
    logging.info("Libraries:")
    logging.info("  elftools: {}".format(elftools.__version__))

    extern_tools_verify()


class ExternTool:
    def __init__(self, cmd, os_pkg, stderr_ignore=[], encoding=None):
        self.cmd = cmd
        self.os_pkg = os_pkg
        self.stderr_ignore = stderr_ignore
        if encoding is None:
            # Same as subprocess.run()
            encoding = locale.getpreferredencoding(False)
        self.encoding = encoding

    def __should_ignore(self, line):
        for ignore in self.stderr_ignore:
            if ignore in line:
                return True
        return False

    def run(self, *args, _internal=False, **kw):
        args = list(args)
        args.insert(0, self.cmd)

        if not _internal:
            logging.debug("Running " + str(args))
        try:
            r = subprocess.run(
                args = args,

                #capture_output = True,         # TODO: Python 3.7
                stdout = subprocess.PIPE,
                stderr = subprocess.PIPE,

                **kw)
        except FileNotFoundError:
            raise MissingToolError(self.cmd, self.os_pkg)

        # TODO: Python 3.6: Simply set encoding in run() call
        r.stdout = r.stdout.decode(self.encoding)
        r.stderr = r.stderr.decode(self.encoding)

        # Hide ignored lines from stderr
        if not _internal:
            for line in r.stderr.splitlines(True):
                if self.__should_ignore(line):
                    continue
                sys.stderr.write(line)

        return r.returncode, r.stdout


    def run_check(self, *args, **kw):
        rc, stdout = self.run(*args, **kw)

        if rc != 0:
            raise ToolError(self.cmd, '{} returned {}'.format(self.cmd, rc))

        return stdout

    def get_version(self):
        rc, output = self.run('--version', _internal=True)
        if rc == 0:
            return output.splitlines()[0]
        return "??? (exited {})".format(rc)

    def which(self):
        return which_exec(self.cmd)



tool_ldd        = ExternTool(os.getenv("STATICX_LDD", "ldd"), 'libc-bin')
tool_objcopy    = ExternTool('objcopy', 'binutils')
tool_patchelf   = ExternTool('patchelf', 'patchelf',
                    stderr_ignore = [
                        'working around a Linux kernel bug by creating a hole',
                    ],
                    # They literally have "e2 80 98" in their source file
                    encoding='utf-8',
                    )
tool_strip      = ExternTool('strip', 'binutils')

all_tools = (tool_ldd, tool_objcopy, tool_strip, tool_patchelf)

def extern_tools_verify():
    logging.debug("External tools:")
    for t in all_tools:
        logging.info("  {}: {}: {}".format(t.cmd, t.which(), t.get_version()))


class LddError(ToolError):
    def __init__(self, message):
        super().__init__('ldd', message)


def _parse_ldd_output(output):
    # Example:
    #	linux-vdso.so.1 (0x00007ffe53551000)
    #     or
    #	linux-vdso.so.1 =>  (0x00007ffe53551000)
    #	libc.so.6 => /usr/lib64/libc.so.6 (0x00007f42ac010000)
    #	/lib64/ld-linux-x86-64.so.2 (0x0000557376e75000)
    #     or
    #   /lib64/ld-linux-x86-64.so.2 => /usr/lib64/ld-linux-x86-64.so.2 (0x00007f1de63ac000)
    pat = re.compile(r'\t([\w./+-]*) (?:=> ([\w./+-]*) )?\((0x[0-9a-fA-F]*)\)')

    for line in output.splitlines():
        m = pat.match(line)
        if not m:
            # Some shared objs might have no DT_NEEDED tags (see issue #67)
            if line == '\tstatically linked':
                break
            raise LddError("Unexpected line in ldd output: " + line)
        libname  = m.group(1)
        resolved = m.group(2)
        baseaddr = int(m.group(3), 16)

        if resolved:
            # ldd outupt a resolved symlink, use that
            libpath = resolved
        elif libname.startswith('/'):
            # The library directly referenced an absolute path, use that
            libpath = libname
        else:
            # Assume an unresolved non-absolute library name is synthetic
            # like linux-vdso*.so or linux-gate.so which should be ignored
            # TODO: This check could be removed/relaxed
            if not libname.startswith('linux-'):
                raise LddError("Unexpected unresolved libname: " + libname)
            logging.debug("Ignoring synthetic library: " + libname)
            continue

        # The library path must be absolute
        if not libpath.startswith('/'):
            raise LddError("Unexpected non-absolute libpath: " + libpath)
        yield libpath


def get_shobj_deps(path, libpath=None):
    """Discover the dependencies of a shared object (*.so file)
    """

    # First verify we're dealing with a dynamic ELF file
    ensure_dynamic(path)

    # Prepare the environment
    keep_vars = {'LD_LIBRARY_PATH'}
    env = {k:v for k,v in os.environ.items() if k in keep_vars}

    if libpath:
        # Prepend to LD_LIBRARY_PATH
        assert isinstance(libpath, list)
        old_libpath = env.get('LD_LIBRARY_PATH', '')
        env['LD_LIBRARY_PATH'] = ':'.join(libpath + [old_libpath])

    rc, output = tool_ldd.run(path, env=env)

    if rc != 0:
        # There are multiple ways this can happen:
        #
        # $ ldd ./static
        #     not a dynamic executable
        #
        # We avoid this case by ensuring we're looking at a dynamic ELF.
        #
        #
        # $ ldd ./dynamic-musl
        # ./dynamic-musl: error while loading shared libraries: /lib64/libc.so: invalid ELF header
        #
        # GNU libc doesn't like something about the ELF headers of object files
        # produced by musl-libc
        #
        # We simply raise a specific exception and let the caller deal with it.
        message = "Unexpected ldd error ({}):\n{}".format(rc, output.strip('\n'))
        if "invalid ELF header" in output:
            message += "\nHint: try setting STATICX_LDD to the appropriate ldd for this executable"
        raise LddError(message)

    return list(_parse_ldd_output(output))



def elf_add_section(elfpath, secname, secfilename):
    tool_objcopy.run_check(
        '--add-section', '{}={}'.format(secname, secfilename),
        elfpath)


def elf_dump_section(elfpath, secname, outpath):
    # https://stackoverflow.com/a/3925113/119527
    tool_objcopy.run_check(
        '-O', 'binary',
        '--only-section', secname,
        '--set-section-flags', '{}={}'.format(secname, 'alloc'),
        elfpath, outpath)



def patch_elf(path, interpreter=None, rpath=None, force_rpath=False, no_default_lib=False, add_needed=None):
    args = []
    if interpreter:
        args += ['--set-interpreter', interpreter]
    if rpath:
        args += ['--set-rpath', rpath]
    if force_rpath:
        # There is a bug in patchelf that requires --remove-rpath to be used
        # first before --force-rpath is effective.
        # https://github.com/NixOS/patchelf/issues/94
        # This was fixed in v0.11 but that's newer than Debian Buster.
        remove_rpath(path)

        args.append('--force-rpath')
    if add_needed:
        for lib in coerce_sequence(add_needed):
            args += ['--add-needed', lib]
    args.append(path)

    tool_patchelf.run_check(*args)

    # There is a bug in patchelf that requires the --force-rpath and
    # --no-default-lib steps to be run in separate invocations.
    # https://github.com/NixOS/patchelf/issues/223
    if no_default_lib:
        tool_patchelf.run_check('--no-default-lib', path)

def remove_rpath(path):
    tool_patchelf.run_check('--remove-rpath', path)

def strip_elf(path):
    tool_strip.run_check(path)


################################################################################
# Using pyelftools

class StaticELFError(Error):
    """Dynamic operation requested on static executable"""
    def __init__(self, path):
        message = "{} is a static ELF file".format(path)
        super().__init__(message)


class ELFFileX(ELFFile):
    def __init__(self, stream, path=None):
        self.__path = path
        super().__init__(stream)

    @classmethod
    def open(cls, path, mode='rb'):
        return cls(open(path, mode), path=path)

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.stream.close()

    def get_single_section(self, sectype):
        """Returns the only section of a given type, or None if absent"""
        def key(sec):
            return isinstance(sec, sectype)
        return single(self.iter_sections(), key=key, default=None)

    def get_prog_interp(self):
        for seg in self.iter_segments():
            # Amazingly, this is slightly faster than
            # if isinstance(seg, InterpSegment):
            try:
                return seg.get_interp_name()
            except AttributeError:
                continue

        raise InvalidInputError("{}: not a dynamic executable "
                                "(no interp segment)".format(self.__path))


    def get_dynamic_segment(self):
        for seg in self.iter_segments():
            if seg['p_type'] == 'PT_DYNAMIC':
                assert isinstance(seg, DynamicSegment)
                return seg
        return None

    def is_dynamic(self):
        return bool(self.get_dynamic_segment())


    def get_single_dynamic_tag(self, name):
        dyn = self.get_dynamic_segment()
        if dyn:
            return single(dyn.iter_tags(name), default=None)
        return None

    def get_rpath(self):
        """Returns the value of the DT_RPATH tag of the ELF file"""
        return self.get_single_dynamic_tag('DT_RPATH')

    def get_runpath(self):
        """Returns the value of the DT_RUNPATH tag of the ELF file"""
        return self.get_single_dynamic_tag('DT_RUNPATH')


def open_elf(path, mode='rb'):
    try:
        return ELFFileX.open(path, mode)
    except ELFError as e:
        raise InvalidInputError("{}: Invalid ELF image: {}".format(path, e))


def get_machine(path):
    with open_elf(path) as elf:
        return elf['e_machine']

def get_prog_interp(path):
    with open_elf(path) as elf:
        return elf.get_prog_interp()

def is_dynamic(path):
    with open_elf(path) as elf:
        return elf.is_dynamic()

def ensure_dynamic(path):
    if not is_dynamic(path):
        raise StaticELFError(path=path)



# StaticX
# Copyright 2017 Jonathon Reinhart
# https://github.com/JonathonReinhart/staticx
#
import shutil
from tempfile import NamedTemporaryFile, mkdtemp
import os
from os.path import basename, islink
import logging
import subprocess

from .errors import *
from .utils import *
from .elf import *
from .archive import SxArchive
from .assets import copy_asset_to_tempfile
from .constants import *
from .hooks import run_hooks
from .version import __version__


class StaticxGenerator:
    """StaticxGenerator is responsible for producing a staticx-ified executable.
    """

    def __init__(self, prog, strip=False, compress=True, debug=False, cleanup=True):
        """
        Parameters:
        prog:   Dynamic executable to staticx
        debug:  Run in debug mode (use debug bootloader)
        """
        self.orig_prog = prog
        self.strip = strip
        self.compress = compress
        self.debug = debug
        self.cleanup = cleanup

        self._generate_called = False
        self._added_libs = {}

        # Temporary output file (bootloader copy)
        self.tmpoutput = None
        self.tmpprog = None
        self.tmpdir = None

        f = NamedTemporaryFile(prefix='staticx-archive-', suffix='.tar')
        self.sxar = SxArchive(fileobj=f, mode='w', compress=self.compress)


    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        if self.cleanup:
            self._cleanup()

    def _cleanup(self):
        if self.tmpoutput:
            os.remove(self.tmpoutput)
            self.tmpoutput = None

        if self.tmpprog:
            os.remove(self.tmpprog)
            self.tmpprog = None

        if self.tmpdir:
            shutil.rmtree(self.tmpdir)
            self.tmpdir = None

        if self.sxar:
            self.sxar.close()
            self.sxar = None


    def _get_bootloader(self):
        # Get a temporary copy of the bootloader
        fbl = copy_asset_to_tempfile('bootloader', self.debug,
                                     prefix='staticx-output-', delete=False)
        with fbl:
            self.tmpoutput = bootloader = fbl.name
        make_executable(bootloader)

        # Verify the bootloader machine matches that of the user program
        bldr_mach = get_machine(bootloader)
        prog_mach = get_machine(self.orig_prog)
        if bldr_mach != prog_mach:
            raise FormatMismatchError("Bootloader machine ({}) doesn't match "
                    "program machine ({})".format(bldr_mach, prog_mach))

        # Run the bootloader for identification
        r = subprocess.run(
                args = [bootloader],
                env = dict(STATICX_BOOTLOADER_IDENTIFY='1'),
                stderr = subprocess.PIPE,
                universal_newlines = True,  # TODO: 'text' in Python 3.7
                )
        r.check_returncode()
        lines = (line.split(':', 1)[1].strip() for line in r.stderr.splitlines())
        logging.debug("Bootloader: " + " ".join(lines))


    def generate(self, output):
        """Generate a Staticx program

        Parameters:
        output: Path where output file is written
        """
        # Only allow generate() to be called once per instance.
        # In the future we might relax this, but YAGNI for now.
        if self._generate_called:
            raise InternalError("generate() already called")
        self._generate_called = True

        # Work on a temp copy of the bootloader which becomes the output program
        self._get_bootloader()

        # First, learn things about the original program
        orig_interp = get_prog_interp(self.orig_prog)
        logging.info("Program interpreter: " + orig_interp)

        # Now modify a copy of the user prog
        self.tmpprog = copy_to_tempfile(self.orig_prog, prefix='staticx-prog-', delete=False).name
        self._fixup_prog()

        if self.strip:
            # TODO: Now that we have ship separate debug/release bootloaders
            # do we ever want and need to do this at staticx-time?
            logging.info("Stripping bootloader {}".format(self.tmpoutput))
            strip_elf(self.tmpoutput)


        # Build the archive to be appended
        self.tmpdir = mkdtemp(prefix='staticx-archive-')
        with self.sxar as ar:
            run_hooks(self)

            ar.add_program(self.tmpprog, basename(self.orig_prog))
            ar.add_interp_symlink(orig_interp)

            # Add all of the libraries
            for libpath in get_shobj_deps(self.orig_prog):
                self.add_library(libpath, exist_ok=True)

        # errr...
        arf = self.sxar.fileobj
        arf.flush()

        # Starting from the bootloader, append archive
        elf_add_section(self.tmpoutput, ARCHIVE_SECTION, arf.name)

        # Move the temporary output file to its final place
        move_file(self.tmpoutput, output)
        self.tmpoutput = None


    def add_library(self, libpath, exist_ok=False):
        """Add a library to the archive

        The library will be added with its base name.
        Symlinks will also be added and followed.
        """
        # See if we've already handled this library
        libname = basename(libpath)
        if libname in self._added_libs:
            if exist_ok:
                return
            raise LibExistsError(libname)

        logging.info("Processing library {} ({})".format(libname, libpath))

        # Handle symlinks
        libpath = self._handle_lib_symlinks(libpath)

        # We're left with a real file at this point
        assert not islink(libpath)

        # Lazily make a copy of the library before modifying it
        def work_on_copy():
            nonlocal libpath

            tmplib = os.path.join(self.tmpdir, basename(libpath))
            logging.info("Copying {} to {}".format(libpath, tmplib))
            shutil.copy(libpath, tmplib)
            libpath = tmplib

            nonlocal work_on_copy
            def work_on_copy(): pass


        # Audit library to check for problems
        try:
            self.check_library_rpath(libpath)
        except (UnsupportedRpathError, UnsupportedRunpathError):
            # Fix it by removing
            work_on_copy()
            logging.info("Removing RPATH/RUNPATH from library {}".format(libpath))
            remove_rpath(libpath)

        # Strip the library
        if self.strip:
            work_on_copy()
            logging.info("Stripping library {}".format(libpath))
            strip_elf(libpath)


        # Finally, add it to the archive.
        arcname = basename(libpath)
        logging.info("Adding {} as {}".format(libpath, arcname))
        self.sxar.add_file(libpath, arcname=arcname)
        if arcname in self._added_libs:
            raise InternalError("libname {} absent from _added_libs but"
                    " library {} present".format(libname, arcname))
        self._added_libs[arcname] = libpath


    def _handle_lib_symlinks(self, libpath):
        """Recursively process symlinks along path to library

        Generate local symlinks inside the archive reflecting these.

        Return the fully-resolved path to the actual library.
        """
        while islink(libpath):
            arcname = basename(libpath)
            libpath = get_symlink_target(libpath)

            # Add a symlink.
            # At this point the target probably doesn't exist, but that doesn't matter yet.
            logging.info("Adding Symlink {} => {}".format(arcname, basename(libpath)))
            self.sxar.add_symlink(arcname, basename(libpath))
            if arcname in self._added_libs:
                raise InternalError("libname {} absent from _added_libs but"
                        " symlink {} present".format(libname, arcname))
            self._added_libs[arcname] = None    # Don't care about real target for symlinks

        return libpath


    def check_library_rpath(self, path, dangerous_only=False):
        """Inspect a library to see if it uses problematic RPATH/RUNPATH

        See https://github.com/JonathonReinhart/staticx/issues/172
        """
        def is_dangerous(rpath):
            # rpath can be:
            # * Absolutute                  dangerous
            # * Relative (to working dir)   dangerous (and stupid)
            # * Relative (to $ORIGIN)       safe (as long as no ..)
            if not rpath.startswith('$ORIGIN'):
                return True

            # There might be some odd corner cases here, but this
            # conservative approach should be good enough.
            if '..' in rpath:
                return True

            return False

        with open_elf(path) as elf:
            # Check for RPATH
            rp = elf.get_rpath()
            if rp and not dangerous_only and is_dangerous(rp.rpath):
                raise UnsupportedRpathError(path, rp.rpath)

            # Check for RUNPATH
            rp = elf.get_runpath()
            if rp:
                # RUNPATH is always dangerous because it kills RPATH
                raise UnsupportedRunpathError(path, rp.runpath)



    def _fixup_prog(self):
        """Fixup our temporary copy of the user's program"""

        # Strip user prog before modifying it
        if self.strip:
            logging.info("Stripping prog {}".format(self.tmpprog))
            strip_elf(self.tmpprog)

        # Set long dummy INTERP and RPATH in the executable to allow plenty of space
        # for bootloader to patch them at runtime, without the reording complexity
        # that patchelf has to do.
        new_interp = 'i' * MAX_INTERP_LEN
        new_rpath = 'r' * MAX_RPATH_LEN
        patch_elf(self.tmpprog, interpreter=new_interp, rpath=new_rpath,
                  force_rpath=True, no_default_lib=True)


def generate(prog, output, libs=None, strip=False, compress=True, debug=False):
    """Main API: Generate a staticx executable

    Parameters:
    prog:   Dynamic executable to staticx
    output: Path to result
    libs: Extra libraries to include
    strip: Strip binaries to reduce size
    debug: Run in debug mode (use debug bootloader)
    """

    logging.info("Running StaticX version {}".format(__version__))
    verify_tools()
    logging.debug("Arguments:")
    logging.debug("  prog:      {!r}".format(prog))
    logging.debug("  output:    {!r}".format(output))
    logging.debug("  libs:      {!r}".format(libs))
    logging.debug("  strip:     {!r}".format(strip))
    logging.debug("  compress:  {!r}".format(compress))
    logging.debug("  debug:     {!r}".format(debug))

    gen = StaticxGenerator(
            prog=prog,
            strip=strip,
            compress=compress,
            debug=debug,
            )
    with gen:
        for lib in (libs or []):
            gen.add_library(lib)

        gen.generate(output=output)



#!/usr/bin/env python
# Derived from https://github.com/JonathonReinhart/scuba
from setuptools import setup, Command, find_packages
from wheel.bdist_wheel import bdist_wheel
from distutils.command.build import build
import os
import sys
from subprocess import check_call


################################################################################
# Dynamic versioning

def get_dynamic_version():
    import staticx.version

    # Travis builds
    # If we're not building for a tag, then append the build number
    build_num = os.getenv('TRAVIS_BUILD_NUMBER')
    build_tag = os.getenv('TRAVIS_TAG')
    if (not build_tag) and (build_num != None):
        return '{}.{}'.format(staticx.version.BASE_VERSION, build_num)

    return staticx.version.__version__


################################################################################
# Commands / hooks

class build_bootloader(Command):
    description = "Build staticx bootloader binary"

    user_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass

    def run(self):
        args = [
            'scons',
            '-Q',       # Quiet output (except for build status)
            'STATICX_VERSION={}'.format(get_dynamic_version()),
        ]
        check_call(args)


class build_hook(build):
    def run(self):
        self.run_command('build_bootloader')
        build.run(self)


class bdist_wheel_hook(bdist_wheel):
    def finalize_options(self):
        self.plat_name = get_platform()     # Equivalent to passing --plat-name
        super().finalize_options()


class UnsupportedPlatformError(Exception):
    pass

def get_platform():
    """Get platform string (e.g. "manylinux1_x86_64")

    Staticx itself is a pure python package, but it includes binary assets
    which are OS/machine dependent. Thus, staticx wheels shouldn't be tagged
    with platform of "any", which is the default behavior when setup.py doesn't
    include any Python extensions.

    There are various ways to change this behavior, none of which works well
    for staticx:
    1) Pass --plat-name to `setup.py bdist_wheel`.
       This gets in the way of external tooling (e.g. Travis 'deploy')
    2) Include a dummy Extension: https://stackoverflow.com/a/53463910
       This causes undue problems becuase it sets the Python ABI field also,
       which staticx doesn't care about.

    So instead, we take care of it ourselves.

    While staticx isn't really built in the typical 'manylinux1' fashion,
    it identifies as 'manylinux1' for PyPI binary wheel compatibility.

    Rationale: staticx assets follow stricter compatibility rules than
    manylinux1:
    - bootloader is statically-linked
    - libnsskill.so is GLIBC-specific and version-agnostic
    """
    uname = os.uname()

    # Apply the same conversions as distutils.util.get_platform()
    osname = uname.sysname.lower().replace('/', '')
    machine = uname.machine.replace(' ', '_').replace('/', '-')

    # Check compatibility
    if osname != "linux":
        # Right now staticx only supports Linux
        raise UnsupportedPlatformError("OS {!r} unsupported by staticx".format(osname))

    # Adjust machine if necessary
    #
    # This applies to a 32-bit Python interpreter running on a 64-bit kernel.
    # We assume that the compiler is also targetting 32-bit, otherwise Python
    # won't match the UINTPTR_MAX check in bootloader/elfutil.h
    #
    # See:
    #   https://github.com/pypa/pip/pull/3497#discussion_r54870308
    #   https://github.com/pypa/wheel/commit/b9dbb6fa257c
    if machine == "x86_64" and sys.maxsize == 0x7fffffff:
        machine = "i686"

    return 'manylinux1_{}'.format(machine)

################################################################################

def read_project_file(path):
    proj_dir = os.path.dirname(__file__)
    path = os.path.join(proj_dir, path)
    with open(path, 'r') as f:
        return f.read()

setup(
    name = 'staticx',
    version = get_dynamic_version(),
    description = 'Build static self-extracting app from dynamic executable',
    python_requires='>=3.5',
    long_description = read_project_file('README.md'),
    long_description_content_type = 'text/markdown',
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Build Tools',
    ],
    license = 'GPL v2 with special exception allowing StaticX to build and'
              ' distribute non-free programs',
    author = 'Jonathon Reinhart',
    author_email = 'jonathon.reinhart@gmail.com',
    url = 'https://github.com/JonathonReinhart/staticx',
    packages = find_packages(),
    package_data = {
        'staticx': ['assets/*/*'],
    },

    # Ugh.
    # https://github.com/JonathonReinhart/staticx/issues/22
    # https://github.com/JonathonReinhart/scuba/issues/77
    # https://github.com/pypa/setuptools/issues/1064
    include_package_data = True,

    zip_safe = False,   # http://stackoverflow.com/q/24642788/119527
    entry_points = {
        'console_scripts': [
            'staticx = staticx.__main__:main',
            'sx-extract = staticx.extract:main',
        ]
    },
    install_requires = [
        'pyelftools',
    ],

    # http://stackoverflow.com/questions/17806485
    # http://stackoverflow.com/questions/21915469
    # PyInstaller setup.py
    cmdclass = {
        'build_bootloader': build_bootloader,
        'build':            build_hook,
        'bdist_wheel':      bdist_wheel_hook,
    },
)
